"""Utilities for filesystem access."""
import shutil


def rmtree(path: str) -> None:
    """Remove a filesystem tree. Ignore failures if the tree doesn't exist."""
    try:
        shutil.rmtree(path)
    except FileNotFoundError:
        pass



"""
A class that builds models with hyperparameters.

A :py:class:`ModelTemplate` is a description of how to create a
compiled machine learning model and the hyperparameters that the
model depends on.

Subclass :py:class:`ModelTemplate` to describe the *architecture*
of your model and the hyperparameters that are used to construct
the model (and the model's optimizer).

Then, pass an instance of your :py:class:`ModelTemplate` subclass
to as subclass of :py:class:`~scalarstop.model.Model` to train an
instance of a machine learning model created from your
:py:class:`ModelTemplate`.

>>> import tensorflow as tf
>>> import scalarstop as sp
>>>
>>> class small_dense_10_way_classifier_v1(sp.ModelTemplate):
...    @sp.dataclass
...
...    class Hyperparams(sp.HyperparamsType):
...        hidden_units: int
...        optimizer: str = "adam"
...
...    def new_model(self):
...        model = tf.keras.Sequential(
...            layers=[
...                tf.keras.layers.Flatten(input_shape=(28, 28)),
...                tf.keras.layers.Dense(
...                    units=self.hyperparams.hidden_units,
...                    activation="relu",
...                ),
...                tf.keras.layers.Dense(units=10)
...           ],
...            name=self.name,
...        )
...        model.compile(
...            optimizer=self.hyperparams.optimizer,
...            loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True),
...            metrics=["accuracy"],
...        )
...        return model

>>> model_template = small_dense_10_way_classifier_v1(hyperparams=dict(hidden_units=20))

>>> model_template.name
'small_dense_10_way_classifier_v1-zc9r3do1baeeffafanjnjmou'

"""
from typing import Any, Mapping, Optional, Union

from scalarstop.exceptions import IsNotImplemented
from scalarstop.hyperparams import HyperparamsType, hash_hyperparams, init_hyperparams


class ModelTemplate:
    """Describes machine learning model architectures and hyperparameters. Used to generate new machine learning model objects that are passed into :py:class:`~scalarstop.model.Model` objects."""  # pylint: disable=line-too-long

    Hyperparams: type
    hyperparams: HyperparamsType

    _name: Optional[str] = None
    _group_name: Optional[str] = None
    _model = None

    def __init__(
        self, *, hyperparams: Optional[Union[Mapping[str, Any], HyperparamsType]] = None
    ):
        self.hyperparams = init_hyperparams(
            class_name=self.__class__.__name__,
            hyperparams=hyperparams,
            hyperparams_class=self.Hyperparams,
        )

    def __repr__(self) -> str:
        return f"<sp.ModelTemplate {self.name}>"

    @property
    def name(self) -> str:
        """
        The name of the specific model with values as hyperparameters.

        If you intend on overriding this method, make sure that
        instances of the same class with different hyperparameters
        will have different names.

        However, if you use additional parameters to your
        :py:class:`~scalarstop.model_template.ModelTemplate` 's
        ``__init__()`` method that are
        *not* hyperparameters--such as paths on your filesystem--then
        you should be sure that changes in those values do *not*
        change your :py:class:`ModelTemplate` name.
        """
        if self._name is None:
            self._name = "-".join((self.group_name, hash_hyperparams(self.hyperparams)))
        return self._name

    @property
    def group_name(self) -> str:
        """
        The group name of this model template.

        Conceptually, the group name is the name for all compiled machine
        learning models that share the same code but have different
        hyperparameters.
        """
        if self._group_name is None:
            self._group_name = self.__class__.__name__
        return self._group_name

    def new_model(self) -> Any:
        """
        Create a new compiled model with the current hyperparameters.

        When you override this method, make sure to create a new model
        object every single time this function is called.
        """
        raise IsNotImplemented("ModelTemplate.new_model()")



"""
A wrapper for cloudpickle to (de)serialize Python objects.
"""

from typing import Any, BinaryIO, TextIO, Union, cast

import cloudpickle

_PICKLE_PROTOCOL_VERSION = 4

_PICKLE_FIX_IMPORTS = True

_PICKLE_ENCODING = "ASCII"

_PICKLE_ERRORS = "strict"


def load(file: Union[BinaryIO, TextIO]) -> Any:
    """Load a Python object from a file handle."""
    return cloudpickle.load(
        file=file,
        fix_imports=_PICKLE_FIX_IMPORTS,
        encoding=_PICKLE_ENCODING,
        errors=_PICKLE_ERRORS,
    )


def loads(data: str) -> Any:
    """Load a Python object from a string."""
    return cloudpickle.loads(
        data=data,
        fix_imports=_PICKLE_FIX_IMPORTS,
        encoding=_PICKLE_ENCODING,
        errors=_PICKLE_ERRORS,
    )


def dump(*, obj: Any, file: Union[BinaryIO, TextIO]) -> None:
    """Dump a Python object to a file handle."""
    cloudpickle.dump(
        obj=obj,
        file=file,
        protocol=_PICKLE_PROTOCOL_VERSION,
    )


def dumps(obj: Any) -> str:
    """Dump a Python object to a string."""
    dumped = cloudpickle.dumps(
        obj=obj,
        protocol=_PICKLE_PROTOCOL_VERSION,
    )
    return cast(str, dumped)



"""Issue warnings."""
import warnings as _python_warnings


class ScalarStopWarning(UserWarning):
    """Parent class for all ScalarStop warnings."""


def warn(message: str) -> None:
    """Issue a ScalarStop warning."""
    _python_warnings.warn(message=message, category=ScalarStopWarning)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scalarstop']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy>=1.4.5,<2.0.0',
 'cloudpickle>=1.6.0,<2.0.0',
 'numpy',
 'pandas',
 'psycopg2-binary>=2.8.6,<3.0.0',
 'tensorflow>=2.3.0']

setup_kwargs = {
    'name': 'scalarstop',
    'version': '0.2.0',
    'description': 'A framework for managing machine learning experiments.',
    'long_description': '.. image:: https://static.neocrym.com/images/scalarstop/v1/1x/scalarstop-wordmark-color-black-on-white--1x.png\n\nKeep track of your machine learning experiments with ScalarStop.\n================================================================\n\nScalarStop is a Python framework for reproducible machine learning research.\n\nIt was written and open-sourced at `Neocrym <https://www.neocrym.com>`_, where it is used to train thousands of models every week.\n\nScalarStop can help you:\n\n* organize datasets and models with *content-addressable* names.\n* save/load datasets and models to/from the filesystem.\n* record hyperparameters and metrics to a relational database.\n\nSystem requirements\n-------------------\nScalarStop is a Python package that requires Python 3.8 or newer.\n\nCurrently, ScalarStop only supports tracking\n`tf.data.Dataset <https://www.tensorflow.org/api_docs/python/tf/data/Dataset>`_\ndatasets and `tf.keras.Model <https://www.tensorflow.org/api_docs/python/tf/keras/Model>`_\nmodels. As such, ScalarStop requires TensorFlow 2.3.0 or newer.\n\nWe encourage anybody that would like to add support for other\nmachine learning frameworks to ScalarStop. :)\n\nInstallation\n------------\n\nScalarStop is `available on PyPI <https://pypi.org/project/scalarstop/>`_.\nYou can install by running the command:\n\n.. code:: bash\n\n    python3 -m pip install scalarstop\n\nIf you would like to make changes to ScalarStop, you can `clone the repository <https://github.com/scalarstop/scalarstop>`_\nfrom GitHub.\n\n.. code:: bash\n\n    git clone https://github.com/scalarstop/scalarstop.git\n    cd scalarstop\n    python3 -m pip install .\n\nUsage\n-----\n\nRead the `ScalarStop Tutorial <https://github.com/scalarstop/scalarstop/blob/main/notebooks/tutorial.ipynb>`_ to learn the core concepts behind ScalarStop and how to structure your datasets and models.\n\nAfterwards, you might want to dig deeper into the `ScalarStop Documentation <https://docs.scalarstop.com>`_. In general, a typical ScalarStop workflow involves four steps:\n\n1. Organize your datasets with `scalarstop.datablob <https://www.scalarstop.com/en/latest/autoapi/scalarstop/datablob/#module-scalarstop.datablob>`_.\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n2. Describe your machine learning model architectures using `scalarstop.model_template <https://www.scalarstop.com/en/latest/autoapi/scalarstop/model_template/#module-scalarstop.model_template>`_.\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n3. Load, train, and save machine learning models with `scalarstop.model <https://www.scalarstop.com/en/latest/autoapi/scalarstop/model/#module-scalarstop.model>`_.\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n4. Save hyperparameters and training metrics to a SQLite or PostgreSQL database using `scalarstop.train_store <https://www.scalarstop.com/en/latest/autoapi/scalarstop/train_store/#module-scalarstop.train_store>`_.\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\nContributing to ScalarStop\n--------------------------\n\nWe warmly welcome contributions to ScalarStop. Here are the technical details for getting started with adding code to ScalarStop.\n\nGetting started\n^^^^^^^^^^^^^^^\nFirst, clone this repository from GitHub. All development happens on the ``main`` branch.\n\n.. code:: bash\n\n    git clone https://github.com/scalarstop/scalarstop.git\n\nThen, run ``make install`` to install Python dependencies in a Poetry virtualenv.\n\nYou can run ``make help`` to see the other commands that are available.\n\nChecking your code\n^^^^^^^^^^^^^^^^^^\nRun ``make fmt`` to automatically format code.\n\nRun ``make lint`` to run Pylint and MyPy to check for errors.\n\nGenerating documentation\n^^^^^^^^^^^^^^^^^^^^^^^^\nDocumentation is important! Here is how to add to it.\n\nGenerating Sphinx documentation\n"""""""""""""""""""""""""""""""\n\nYou can generate a local copy of our Sphinx documentation at `scalarstop.com <https://www.scalarstop.com/en/latest/>`_ with ``make docs``.\n\nThe generated documentation can be found at ``docs/_build/dirhtml``. To view it, you should start an HTTP server in this directory, such as:\n\n.. code:: bash\n\n    make docs\n    cd docs/_build/dirhtml\n    python3 -m http.server 5000\n\nThen visit http://localhost:5000 in your browser to preview changes to the documentation.\n\nIf you want to use Sphinx\'s ability to automatically generate hyperlinks to the Sphinx documentation of other Python projects, then you should configure `intersphinx <https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html>`_ settings at the path docs/conf.py. If you need to download an objects.inv file, make sure to update the ``make update-sphinx`` command in the `Makefile <https://github.com/scalarstop/scalarstop/blob/main/Makefile>`_.\n\nEditing the tutorial notebook\n"""""""""""""""""""""""""""""\nThe main ScalarStop tutorial is `in a Jupyter notebook <https://github.com/scalarstop/scalarstop/blob/main/notebooks/tutorial.ipynb>`_. If you have made changes to ScalarStop, you should rerun the Jupyter notebook on your machine with your changes to make sure that it still runs without error.\n\nRunning unit tests\n^^^^^^^^^^^^^^^^^^\nRun ``make test`` to run all unit tests.\n\nIf you want to run a specific unit test, try running ``python3 -m poetry run python -m unittest -k {name of your test}``.\n\nUnit tests with SQLite3\n"""""""""""""""""""""""\nIf you are running tests using a Python interpreter that does not have the `SQLite3 JSON1 extension <https://www.sqlite.org/json1.html>`_, then `TrainStore <https://www.scalarstop.com/en/latest/autoapi/scalarstop/train_store/#module-scalarstop.train_store>`_ unit tests involving SQLite3 will be skipped. This is likely to happen if you are using Python 3.8 on Windows. If you suspect that you are missing the SQLite3 JSON1 extension, the `Django documentation has some suggestions <https://code.djangoproject.com/wiki/JSON1Extension>`_ for how to fix it.\n\nUnit tests with PostgreSQL\n""""""""""""""""""""""""""\nBy default, tests involving PostgreSQL are skipped. To enable PostgreSQL, run ``make test`` in a shell where the environment variable ``TRAIN_STORE_CONNECTION_STRING`` is set to a `SQLAlchemy database connection URL <https://docs.sqlalchemy.org/en/14/core/engines.html>`_--which looks something like ``"postgresql://scalarstop:changeme@localhost:5432/train_store"``. The connection URL should point to a working PostgreSQL database with an existing database and user.\n\nThe docker-compose.yml file in the root of this directory can set up a PostgreSQL instance on your local machine. If you have `Docker <https://docs.docker.com/get-docker/>`_ and `Docker Compose <https://docs.docker.com/compose/install/>`_ installed, you can start the PostgreSQL database by running ``docker-compose up`` in the same directory as the docker-compose.yml file.\n\nMeasuring test coverage\n"""""""""""""""""""""""\nYou can run ``make test-with-coverage`` to collect Python line and branch coverage information. Afterwards, run ``make coverage-html`` to generate an HTML report of unit test coverage. You can view the report in a web browser at the path ``htmlcov/index.html``.\n\nCredits\n-------\nScalarStop\'s documentation is built with `Sphinx <https://www.sphinx-doc.org/>`_ using `@pradyunsg <https://pradyunsg.me>`_\'s `Furo <https://github.com/pradyunsg/furo>`_ theme and is hosted by `Read the Docs <https://readthedocs.org/>`_.\n',
    'author': 'Neocrym Records Inc',
    'author_email': 'engineering@neocrym.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.scalarstop.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Every custom ScalarStop exception.
"""

from typing import Any

from scalarstop.dataclasses import fields


class ScalarStopException(Exception):
    """
    Base class for all ScalarStop exceptions.
    """


class YouForgotTheHyperparams(ValueError, ScalarStopException):
    """Raised when the user creates a class, but forgets to add a nested Hyperparams class."""

    def __init__(self, class_name: str):
        super().__init__(
            "Please make sure to define a dataclass containing "
            f"hyperparameters at `{class_name}.Hyperparams`. "
            "Your dataclass should subclass `sp.HyperparamsType` "
            "and is decorated with a `@sp.dataclass` decorator."
        )


class WrongHyperparamsType(TypeError, ScalarStopException):
    """Raised when the user passes an object with the wrong type as hyperparams."""

    def __init__(self, *, hyperparams: Any, class_name: str):
        super().__init__(
            f"You passed an object with the type {type(hyperparams)} as hyperparams. "
            "Check with the surrounding class. You should either pass a dictionary "
            "or other mapping-like object or directly pass "
            f"a {class_name}.Hyperparams object."
        )


class WrongHyperparamsKeys(TypeError, ScalarStopException):
    """Raised when the user has passed extra or missing keys for constructing hyperparams."""

    def __init__(self, hyperparams: Any, hyperparams_class: type):
        hyperparams_class_fields = [field.name for field in fields(hyperparams_class)]
        super().__init__(
            "Wrong keys passed to create hyperparams. "
            f"Valid keys include {hyperparams_class_fields}. "
            f"The object you passed is: {hyperparams}"
        )


class DataBlobNotFound(FileNotFoundError, ScalarStopException):
    """
    Raised when we cannot load a :py:class:`DataBlob`
    from the filesystem.
    """

    def __init__(self, path: str):
        """
        Args:
            path: The directory that we went
                looking for the :py:class:`DataBlob` when we
                were not able to find it.
        """
        super().__init__("Could not load a DataBlob from the filesystem at " + path)


class ModelNotFoundError(FileNotFoundError, ScalarStopException):
    """Raised when we cannot find a saved model."""

    def __init__(self, directory: str):
        super().__init__(f"Could not find a saved model at {directory}.")


class FileExists(FileExistsError, ScalarStopException):
    """
    Raise when we want to save a :py:class:`DataBlob` but a
    file exists at the save path
    """


class FileExistsDuringDataBlobCreation(FileExists):
    """
    Raised when a file exists at the save path, but we think the
    file was created *while* we were saving the :py:class:`DataBlob`,
    which suggests that the user's code has a race conditon.
    """


class TensorFlowDatasetNotFound(FileNotFoundError, ScalarStopException):
    """
    Raised when we canno load a :py:class:`tf.data.Dataset`
    from the filesystem.
    """

    def __init__(self, directory: str):
        super().__init__(f"Could not load a tf.data.Dataset from {directory}")


class ElementSpecNotFound(FileNotFoundError, ScalarStopException):
    """
    Raised when we try to guess the location of a serialized
    ``element_spec`` for a :py:class:`tf.data.Dataset`.
    """

    def __init__(self, directory: str):
        """
        Args:
            directory: The directory where we were looking for
                the ``element_spec`` for a :py:class:`tf.data.Dataaset`.
                The TensorFlow API assumes that you have the
                ``element_spec`` on hand, but ScalarStop likes to
                save it to disk.

        """
        super().__init__(
            "Could not find an element_spec for the tf.data.Dataset "
            f"at {directory}. Either check your path or provide an "
            "element_spec."
        )


class IsNotImplemented(NotImplementedError, ScalarStopException):
    """
    Raised when someone calls a function that isn't yet implemented.
    """

    def __init__(self, name: str):
        """
        Args:
            name: The name of the function, method, or property that
                was not implemented. This is useful in case the stack
                trace is hard to read for some reason.
        """
        super().__init__(f"Make sure to implement `{name}` in a subclass.")


class SQLite_JSON_ModeDisabled(RuntimeError, ScalarStopException):
    """
    Raised when the user is running the
    :py:class:`~scalarstop.TrainStore` backed by
    a SQLite database, but the SQLite JSON1 extension is
    unavailable.
    """

    def __init__(self) -> None:
        super().__init__(
            "You want to start the TrainStore with a SQLite "
            "database, but we could not find the SQLite JSON1 "
            "extension in your Python installation."
        )
